@@ -1,4 +1,4 @@
-# Copyright (c) 2003-2007 LOGILAB S.A. (Paris, FRANCE).
+# Copyright (c) 2003-2009 LOGILAB S.A. (Paris, FRANCE).
 # http://www.logilab.fr/ -- mailto:contact@logilab.fr
 #
 # This program is free software; you can redistribute it and/or modify it under
@@ -20,3 +20,12 @@ written with efficiency in mind when dealing with big repositories
 (it can happily be used to browse Linux kernel source code
 repository).
 """
+try:
+    from mercurial.version import get_version
+    merc_version = get_version()
+except ImportError:
+    from mercurial.util import version
+    merc_version = version()
+
+print "*"*80
+print "Using Mercurial version", merc_version
