@@ -96,6 +96,12 @@ def make_texttag( name, **kwargs ):
             print "Warning the property %s is unsupported in this version of pygtk" % key
     return tag
 
+class SearchState(object):
+    __slots__ = ('rev_iter', 'text_iter', 'last_grep_string', 'last_find_string')
+    def __init__(self, **kwargs):
+        for k, v in kwargs.iteritems():
+            setattr(self, k, v)
+
 class HgViewApp(object):
     """Main hg view application"""
     def __init__(self, repo, filerex = None ):
@@ -113,6 +119,7 @@ class HgViewApp(object):
             self.filter_files = None
         self.filter_noderange = None
         self.branch_selected = None
+        self._search_state = None
         self.SHOW_GRAPH = True
         # The strings are stored as PYOBJECT when they contain zeros and also
         # to save memory when they are used by the custom renderer
@@ -164,7 +171,7 @@ class HgViewApp(object):
 
     def revpopup_add_tag(self, item):
         raise NotImplementedError()
-        
+
     def revpopup_update(self, item):
         raise NotImplementedError()
 
@@ -177,7 +184,7 @@ class HgViewApp(object):
         for branch_name in self.repo.get_branch().keys():
             self.branch_store.append( (branch_name,) )
         combo_filter.set_active(0)
- 
+
     def get_node_branches(self):
         branch_nodeRev = {}
         dic_branch, _unknown, _tip_rev = self.repo.get_branches_heads()
@@ -186,16 +193,16 @@ class HgViewApp(object):
             rev = node_info.rev
             branch_nodeRev[rev] = branch
         return branch_nodeRev
-        
+
     def on_refresh_activate(self, arg):
         self.repo.refresh()
         self.refresh_tree()
-        
+
     def filter_nodes(self):
         """Filter the nodes according to filter_files and filter_nodes"""
         if not self.filter_files and not self.filter_noderange:
             return self.repo.nodes
-        
+
         keepnodes = []
         nodes = self.repo.nodes
         frex = self.filter_files
@@ -208,14 +215,14 @@ class HgViewApp(object):
                 if hide_otherbranches:
                     if node.branches['branch'] != branch_selected and branch_selected != 'All':
                         continue
-                if not node.files: 
+                if not node.files:
                     keepnodes.append(n)
                 else:
                     for f in node.files:
                         if frex.search(f):
                             keepnodes.append( n )
                             break
-  
+
         tree = self.xml.get_widget( "treeview_revisions" )
         file_filter = self.xml.get_widget("entry_file_filter").get_text()
         if file_filter:
@@ -226,7 +233,7 @@ class HgViewApp(object):
             col = tree.get_column(1)
             self.revisions.set_sort_column_id(M_GRAPHORDER, gtk.SORT_ASCENDING)
             self.SHOW_GRAPH = True
-        
+
         return keepnodes
 
     def on_window_main_delete_event( self, win, evt ):
@@ -245,7 +252,19 @@ class HgViewApp(object):
                          u'David Douard',
                          u'Graziella Toutoungis',
                          ])
-        from hgview.__pkginfo__ import modname, version, short_desc, long_desc
+   
+        try:
+            from __pkginfo__ import short_desc, long_desc, version, modname
+        except ImportError:
+            short_desc = "mercurial interactive history viewer"
+            long_desc = """
+            Its purpose is similar to the hgk tool of mercurial, and it has been
+            written with efficiency in mind when dealing with big repositories
+            (it can happily be used to browse Linux kernel source code
+            repository).
+            """
+            version ='unknown'
+            modname ='hgview' 
         dlg.set_comments(short_desc)
         dlg.set_name(modname)
         dlg.set_version(version)
@@ -258,7 +277,7 @@ class HgViewApp(object):
         textwidget = self.xml.get_widget( "textview_status" )
         text_buffer = textwidget.get_buffer()
         tag_table = text_buffer.get_tag_table()
-       
+
         tag_table.add( make_texttag( "mono", family="Monospace" ))
         tag_table.add( make_texttag( "blue", foreground='blue' ))
         tag_table.add( make_texttag( "red", foreground='red' ))
@@ -280,9 +299,8 @@ class HgViewApp(object):
 
     def link_event( self, tag, widget, event, iter_ ):
         """Handle a click on a 'link' tag in the main TextView"""
-        if event.type == gtk.gdk.MOTION_NOTIFY and not (
-            event.state & (gtk.gdk.BUTTON1_MASK|gtk.gdk.BUTTON2_MASK|gtk.gdk.BUTTON3_MASK)
-            ):
+        if event.type == gtk.gdk.MOTION_NOTIFY and not \
+                (event.state & (gtk.gdk.BUTTON1_MASK|gtk.gdk.BUTTON2_MASK|gtk.gdk.BUTTON3_MASK)):
             # we change cursor only if no button is pressed
             cursor = gtk.gdk.Cursor(gtk.gdk.HAND2)
             widget.get_window(gtk.TEXT_WINDOW_TEXT).set_cursor(cursor)
@@ -335,17 +353,14 @@ class HgViewApp(object):
         node = model.get_value( iter_, M_NODE )
         cell.set_property( "text", node.date )
         reversed = column.get_sort_order() == gtk.SORT_DESCENDING
-
         row = model[iter_]
-   
-         
+
     def files_data_func( self, column, cell, model, iter_, user_data=None ):
         """A Cell datafunction used to provide the files"""
         node = model.get_value( iter_, 0 )
-        
-        txt = self.xml.get_widget( "entry_find" ).get_text()
+        txt = self._get_grep_string()
         str_node = str(node).replace('&', '&amp;').replace('<', '&lt;')
-        if txt !='':
+        if txt:
             try:
                 rexp = re.compile( '(%s)' % txt.replace('&', '&amp;').replace('<', '&lt;') )
                 markup = rexp.sub('<span background="yellow">\\1</span>', str_node)
@@ -355,7 +370,7 @@ class HgViewApp(object):
         else:
             markup = str_node
         cell.set_property("markup", '<span>%s</span>' % markup)
-  
+
     def setup_tree(self ):
         """Configure the 2 gtk.TreeView"""
         # Setup the revisions treeview
@@ -398,7 +413,7 @@ class HgViewApp(object):
         tree.append_column( col )
 
         tree.set_model(self.revisions )
-            
+
         ##Setup the filelist treeview
         tree_files = self.xml.get_widget( "treeview_filelist" )
         #tree_files.set_rules_hint( 1 )
@@ -411,14 +426,14 @@ class HgViewApp(object):
         col.set_resizable(True)
         tree_files.append_column( col )
         tree_files.set_model( self.filelist )
-       
+
         rend = DiffStatRenderer()
         col = gtk.TreeViewColumn("Diff Stat", rend, stats=2 )
         col.set_reorderable(True)
         tree_files.append_column( col )
 
         tree_files.set_model( self.filelist )
-        
+
     def sort_on_id(self, *args):
         sorted_column = args[0].get_sort_column_id()
         sort_type = args[0].get_sort_order().value_nick
@@ -426,11 +441,11 @@ class HgViewApp(object):
             if sorted_column == 5:
                 self.SHOW_GRAPH = True
             else:
-                self.SHOW_GRAPH = False 
+                self.SHOW_GRAPH = False
         else:
             self.SHOW_GRAPH = False
-        return self.SHOW_GRAPH 
-              
+        return self.SHOW_GRAPH
+
     def cell_activated(self, *args):
         print "nudge"
 
@@ -451,7 +466,7 @@ class HgViewApp(object):
         self._idletask_id = gobject.idle_add( self.idle_fill_model )
         return
 
-  
+
     def on_treeview_revisions_button_press_event(self, treeview, event):
         if event.button==3:
             x = int(event.x)
@@ -493,8 +508,8 @@ class HgViewApp(object):
                 continue
             rnode = self.repo.read_node( node )
             lines = graph.rowlines[n]
-            auth = self.repo.authors[rnode.author_id] 
-            authcolor = self.repo.colors[rnode.author_id] 
+            auth = self.repo.authors[rnode.author_id]
+            authcolor = self.repo.colors[rnode.author_id]
             rev = int(rnode.rev)
             add_rev( (node, rnode, graph.x[node], (lines,n), time.mktime(rnode.localtime), n, auth, authcolor, rev) )
 
@@ -519,7 +534,7 @@ class HgViewApp(object):
         buf.insert( eob, "Branch: %s\n" % self.repo.read_node(node).branches['branch'])
         buf.insert( eob, "Author: %s\n" %  repo.authors[rnode.author_id] )
         buf.create_mark( "begdesc", buf.get_start_iter() )
-        
+
         for p in repo.parents(node):
             pnode = repo.read_node(p)
             short = short_hex(p)
@@ -564,14 +579,14 @@ class HgViewApp(object):
                 idx += 1
                 markname = "file%d" % idx
                 statmax = max( statmax, stats[0]+stats[1] )
-                stats = [0,0] 
+                stats = [0,0]
                 filespos.append(( file_added , markname, offset, stats ))
-                offset += len(txt) 
+                offset += len(txt)
         else:
 
             for i,l in enumerate(difflines):
                 if l.startswith("diff"):
-                    f = l.split()[-1]
+                    f = l.split(' ', 5)[-1]
                     DIFFHDR = "=== %s ===\n"
                     txt = DIFFHDR % f
                     addtag( "greybg", offset, len(txt) )
@@ -625,6 +640,15 @@ class HgViewApp(object):
             enddesc.backward_line()
             text_buffer.create_mark( "enddesc", enddesc )
             self.filelist.append( ("Content", "begdesc", None ) )
+
+            if len(self.repo.parents(node)) > 1:
+                node_files=[]
+                nodes_parents = self.repo.parents(node)
+                for node_value in nodes_parents:
+                    parent_files = self.repo.read_node(node_value).files
+                    [node_files.append(x) for x in parent_files]
+                rnode.files = tuple(node_files)
+
             diffmsg, buff, changes = self.repo.diff(self.repo.parents(node), node, rnode.files)
             try:
                 buff = unicode( buff, "utf-8" )
@@ -634,7 +658,7 @@ class HgViewApp(object):
             difflines = buff.splitlines()
             del buff
             eob = text_buffer.get_end_iter()
-            
+
             offset = eob.get_offset()
             fileoffsets, tags, lines, statmax = self.prepare_diff( difflines, offset, diffmsg, changes)
             txt = u"".join(lines)
@@ -648,7 +672,7 @@ class HgViewApp(object):
                 i1 = text_buffer.get_iter_at_offset( p1 )
                 txt = text_buffer.get_text( i0, i1 )
                 text_buffer.apply_tag_by_name( name, i0, i1 )
-                
+
             # inserts the marks
             for f, mark,offset, stats in fileoffsets:
                 pos = text_buffer.get_iter_at_offset( offset )
@@ -659,34 +683,6 @@ class HgViewApp(object):
         sob, eob = text_buffer.get_bounds()
         text_buffer.apply_tag_by_name( "mono", sob, eob )
 
-    def hilight_search_string( self ):
-        ##Highlight the search string on the text
-        textwidget = self.xml.get_widget( "textview_status" )
-        text_buffer = textwidget.get_buffer()
-        if not self.find_text:
-            return
-        try:
-            rexp = re.compile(self.find_text)
-        except:
-            return
-        sob, eob = text_buffer.get_bounds()
-        # RESTRICT TO DESCRIPTION PART : should be conditionalized
-        #mark = text_buffer.get_mark( "enddesc" )
-        #enddesc = text_buffer.get_iter_at_mark(mark)
-        #txt = text_buffer.get_slice(sob, enddesc, True )
-        txt = text_buffer.get_slice(sob, eob)
-        m = rexp.search( txt )
-        max_search = 50
-        while m and max_search > 0: # FIXME wrong coloring in diff body
-            _b = text_buffer.get_iter_at_offset( m.start() )
-            _e = text_buffer.get_iter_at_offset( m.end() )
-            if max_search == 50: 
-                mark = text_buffer.create_mark('mark_search', _b, True)
-                textwidget.scroll_to_mark(mark, .2, use_align=True, xalign=1., yalign=0.1)            
-            text_buffer.apply_tag_by_name("yellowbg", _b, _e )
-            m = rexp.search( txt, m.end() )
-            max_search -= 1
-            
     def find_entry(self):
         # text to find
         find_entry = self.xml.get_widget( "entry_find" )
@@ -702,35 +698,31 @@ class HgViewApp(object):
         tw.scroll_to_mark( mark, .2, use_align=True, xalign=1., yalign=0. )
 
     def find_next_row( self, iter, stop_iter=None ):
-        """Find the next revision row based on the content of
-        the 'entry_find' widget"""
-        txt = self.xml.get_widget( "entry_find" ).get_text()
-        rexp = re.compile( txt )
-        while iter != stop_iter and iter!=None:
+        """Find the next revision row matching grep & find entries"""
+        grep = self._get_grep_string()
+        find = self._get_find_string()
+        grexp = re.compile( grep )
+        frexp = re.compile( find )
+        while iter != stop_iter and iter != None:
             revnode = self.revisions.get( iter, M_NODE ) [0]
             # look in author_id, log
             author = self.repo.authors[revnode.author_id]
-            if ( rexp.search( author ) or
-                 rexp.search( revnode.desc ) ):
+            if ( grexp.search( author ) or
+                 grexp.search( revnode.desc ) ):
                 break
             # look in diffs
             node = self.revisions.get_value(iter, M_ID)
             rnode = self.repo.read_node(node)
             buff = self.repo.single_diff(self.repo.parents(node)[0], node, rnode.files)
-            if rexp.search(buff):
+            if grexp.search(buff):
                 break
-            # look file names
-            for f in revnode.files:
-                if rexp.search( f ):
+            for fname in revnode.files:
+                if frexp.search( fname ):
                     break
-            else:
-                iter = self.revisions.iter_next( iter )
-                continue
-            break
-        if iter==stop_iter or iter is None:
+            iter = self.revisions.iter_next( iter )
+        if iter == stop_iter or iter is None:
             return None
         self.select_row( iter )
-        self.hilight_search_string()
         return iter
 
     def select_row( self, itr ):
@@ -739,7 +731,6 @@ class HgViewApp(object):
             return
         else:
             self.find_text = self.xml.get_widget( "entry_find" ).get_text()
-
         tree = self.xml.get_widget( "treeview_revisions" )
         sel = tree.get_selection()
         sel.select_iter( itr )
@@ -754,32 +745,94 @@ class HgViewApp(object):
             it = model.get_iter_first()
         return model, it
 
+    def _get_grep_string(self):
+        return self.xml.get_widget( "entry_find" ).get_text().replace('(', '\(').replace(')', '\)')
+
+    def _get_find_string(self):
+        return self.xml.get_widget( "entry_file_filter" ).get_text()
+
+    def _get_buffer_text(self):
+        textwidget = self.xml.get_widget( "textview_status" )
+        text_buffer = textwidget.get_buffer()
+        sob, eob = text_buffer.get_bounds()
+        return unicode(text_buffer.get_slice(sob, eob), 'UTF-8')
+
+    def notify_end_of_search(self):
+        self.update_status("end of search for %r, %r" % (self._get_grep_string(),
+                                                         self._get_find_string))
+
     @watch_displayed
     def on_button_find_clicked( self, *args ):
         """callback: clicking on the find button
         makes the search start at the row after the
-        next row
+        current row
         """
-        model, it = self.get_selected_rev()
-        it = self.revisions.iter_next( it )
-        start_it = it
-        res = self.find_next_row( it )
-        if res is None:
-            self.find_next_row( self.revisions.get_iter_first(), start_it )
+        st = self._search_state
+        grep_string = self._get_grep_string()
+        find_string = self._get_find_string()
+        if st is None or st.last_grep_string != grep_string or st.last_find_string != find_string:
+            _, it = self.get_selected_rev()
+            st = self._search_state = SearchState(last_grep_string=grep_string,
+                                                  last_find_string=find_string,
+                                                  rev_iter=self.revisions.iter_next( it ))
+            if st.rev_iter is None: # there was nothing at all
+               self._search_state = None
+               self.notify_end_of_search()
+               return
+            st.text_iter = re.finditer(re.compile('(%s)' % st.last_grep_string.replace('&', '&amp;').replace('<', '&lt;')),
+                                       self._get_buffer_text())
+
+        next_match = None
+        try:
+            next_match = st.text_iter.next()
+        except StopIteration:
+            try:
+                st.rev_iter = self.find_next_row( self.revisions.iter_next( st.rev_iter ))
+            except StopIteration:
+                st.rev_iter = None
+            if st.rev_iter is None:
+                self._search_state = None
+                self.notify_end_of_search()
+                return
+            st.text_iter = re.finditer(re.compile('(%s)' % st.last_grep_string.replace('&', '&amp;').replace('<', '&lt;')),
+                                       self._get_buffer_text())
+            try: 
+                next_match = st.text_iter.next()
+            except StopIteration:
+                return
+
+        # Search State is OK
+        self.highlight_search_string(next_match)
+
+    def highlight_search_string(self, match):
+        textwidget = self.xml.get_widget( "textview_status" )
+        text_buffer = textwidget.get_buffer()
+        _b = text_buffer.get_iter_at_offset( match.start() )
+        _e = text_buffer.get_iter_at_offset( match.end() )
+        mark = text_buffer.create_mark('mark_search', _b, True)
+        textwidget.scroll_to_mark(mark, .1, use_align=True, xalign=1., yalign=0.1)
+        text_buffer.apply_tag_by_name("yellowbg", _b, _e )
+
+    def update_status(self, msg):
+        statusbar = self.xml.get_widget("statusbar1")
+        statusbar.push(True, msg)
 
     def on_entry_find_changed( self, *args ):
-        """callback: each keypress triggers a lookup
-        starting at the current row which allows the
-        highlight string to grow without changing rows"""
-        # DISACTIVATED for now (way too slow on medium/big repos)
-        return
+        """callback: each keypress triggers verification of the search string
+        regular expression syntax"""
+        try:
+            st = self._get_grep_string()
+            re.compile(st)
+        except:
+            self.update_status('Wrong regular expression : %s' % st)
+        else:
+            self.update_status('Your regular expression : %s' % st)
 
     def on_entry_find_activate( self, *args ):
         """Pressing enter in the entry_find field does the
         same as clicking on the Find button"""
         self.on_button_find_clicked()
 
-
     def on_filter1_activate( self, *args ):
         self.filter_dialog.show()
 
@@ -826,7 +879,7 @@ class HgViewApp(object):
         combo_filter = self.xml.get_widget("branch_highlight_combo")
         active_branch = combo_filter.get_active_text()
         return active_branch
-    
+
     def get_value_branch_checkbox(self):
         """
         return True or False
@@ -835,9 +888,11 @@ class HgViewApp(object):
 
 
 def main():
-    from hgview import __pkginfo__
-    VERSION=str(__pkginfo__.numversion)
-    
+    try:
+        import __pkginfo__
+        VERSION=str(__pkginfo__.numversion)
+    except ImportError:
+        VERSION='unknown'
     parser = OptionParser(usage="%prog [-R]", version="%%prog %s" %(VERSION))
     parser.add_option( '-R', '--repository', dest='repo',
                        help='location of the repository to explore' )
@@ -845,8 +900,9 @@ def main():
                        help='filter revisions which touch FILE', metavar="FILE")
     parser.add_option( '-g', '--regexp', dest='filerex',
                        help='filter revisions which touch FILE matching regexp')
-    
+
     opt, args = parser.parse_args()
+
     dir_ = None
     if opt.repo:
         dir_ = opt.repo
